#include "SDPdef.h"
int PreMultGivens(int, int, double, double, double **, int);
int ComputeGivens(double, double, double *, double *);


double VDotDD(double *v1, double *v2, int n)
{ /* Dense vector Dot product */
  int i;
  double ans = 0.0;

  for (i=0;i<n;i++){
    ans += v1[i]*v2[i];
  }
  return ans;
}


int CheckPSD(sdpdat *sdt,
	     double *coef,
	     chfac *sf)
{ /* Check whether the matrix (C - sum coef(i)A_i) is psd */
  /* Return 0 if the matrix is pos def, and return positive otherwise */
  int i,j,k, ncoef=0, ndiag=0, mat2size,nden, n;
  double dtemp;
  smwTools *smw = sdt->smw;

  n = sdt->n;
  nden=smw->ndense;
  mat2size = 2*nden;
  if (n==0) return 0;

  /* Check diagonals of the sparse matrix */
  for (i=0;i<nden;i++){
    smw->ndInd[i]=0;
  }

  for (i=0;i<n;i++){
    if (sf->diag[i]<=0.0){
      if (ndiag<nden){
	smw->ndInd[ndiag]= i;
      }
      ndiag++;
    }
  }

  for (i=0;i<nden;i++){
    if (coef[smw->dind[i]]<0){
      ncoef ++;
    }
  }

  if (ndiag > ncoef) return 1; /* Too many negative diagonal elements */
  if ((ndiag == 0) && (ncoef == nden) ){
    return 0;               /* No negative diagonal elements */
  } 


  /* Add dense constraints */
  for (i=0;i<nden;i++){
    MakeDense(&sdt->amats->rows[smw->dind[i]],sdt->wd,n);
    ForwSubst(sf,sdt->wd,smw->upd[i]);
    for (j=0; j<ndiag; j++){
      smw->upd[i][smw->ndInd[j]] = -(smw->upd[i][smw->ndInd[j]]);
    }
  }
  
  for (i=0;i<mat2size;i++){
    for(j=i;j<mat2size;j++){
      if (i<nden){
        if (j<nden){
	  dtemp=VDotDD(smw->upd[i],smw->upd[j],n);
	  smw->mat2[i][j-i] = dtemp;
        }
	else if (j<nden+ndiag){
	  dtemp = smw->upd[i][smw->ndInd[j-nden]];
	  smw->mat2[i][j-i] = dtemp; 
	}
	else {smw->mat2[i][j-i]=0.0;}
      }
      else{
	if ((i==j)&&(j<nden+ndiag)) {smw->mat2[i][0] = 1.0;}
	else {smw->mat2[i][j-i] = 0.0;}
      }
      for (k=0; k<i ;k++){
        if (k<nden+ndiag){
          smw->mat2[i][j-i] = smw->mat2[i][j-i] - 
	    smw->mat2[k][i-k]*smw->mat2[k][j-k]/smw->mat2[k][0];
	  }
      }
    } /* end for j */
  } /* end for i */

 
  for (i=0;i<mat2size; i++){
    smw->v2[i] = 1.0;
  }
  for (i=0;i<nden+ndiag;i++){
    smw->v2[i] = sqrt( fabs(smw->mat2[i][0]) );
  }
 
  for (i=0;i<mat2size;i++){
    for (j=i;j<mat2size;j++){
      dtemp = 0.0;
      for (k=j;k<nden+ndiag;k++){
 	if (k<nden){
	  dtemp = dtemp +
	    smw->mat2[i][k-i]*smw->mat2[j][k-j]*(-coef[smw->dind[k]]);
	}
	else{
	  dtemp = dtemp + 
	          smw->mat2[i][k-i]*smw->mat2[j][k-j]*(-2);
	}
      }
      smw->mat2[i][j-i]=dtemp /(smw->v2[i]*smw->v2[j]);
    }
  }
  
  for (i=0; i<mat2size; i++){
    smw->mat2[i][0] = smw->mat2[i][0] + 1.0;
  }

  return  ( CholFactor(smw->mat2,mat2size, smw->v2) ); 
}



double LPVariables(sdpdat *sdt, double *v, double *ans)
{
  /* Calculates (ans = cc - amat^T v) and returns positive if ans>0  */
  int i,j, nnz, m=sdt->m, p=sdt->p;
  double *cc = sdt->cvec;
  spvec * spv;

  if (p==0) return 1;

  for (i=0;i<p;i++){
    ans[i]=cc[i];
  }

  for (i=0;i<m;i++){
    spv = &(sdt->avecs->rows[i]);
    nnz = spv->nn0;
    for (j=0; j<nnz; j++){
      ans[spv->ja[j]] -= (spv->an[j])*v[i];
    }
  }

  for (i=0;i<p;i++){
    if (ans[i] <= 0)
      return (ans[i]);
  }

  return 1;

}




int RVectorCoeffs(sdpdat *sdt, double *coef){

  int i,j,k, ndiag=0, mat2size,nden, n,nn;
  double temp, dtemp, *rhs, **coeffmat, **svd, **givenproj,c,s;
  smwTools *smw = sdt->smw;

  n = sdt->n;
  nden=smw->ndense;
  mat2size = 2*nden;
  nn = mat2size;
  if (nn==0) return 0;
  if (n==0) return 0;

  /* Check diagonals of the sparse matrix */
  for (i=0;i<nden;i++){
    smw->ndInd[i]=0;
  }
  for (i=0;i<n;i++){
    if (sdt->sf->diag[i]<0){
      if (ndiag<nden){
	smw->ndInd[ndiag]= i;
      }
      ndiag++;
    }
  }
  sdt->smw->ndiag=ndiag;

  rhs = (double*) malloc(nn*sizeof(double));

  coeffmat = (double **) malloc(nn*sizeof(double *));
  givenproj = (double **) malloc(nn*sizeof(double *));
  svd = (double **) malloc(nn*sizeof(double *));

  coeffmat[0] = (double*) malloc(nn*nn*sizeof(double));
  givenproj[0] = (double*) malloc(nn*nn*sizeof(double));
  svd[0] = (double*) malloc(nn*nn*sizeof(double));

  for (i=0; i<nn*nn; i++){
    coeffmat[0][i]=0.0;
    givenproj[0][i]=0.0;
    svd[0][i]=0.0;
  }

  for (i=1; i<nn; i++){
    coeffmat[i]=coeffmat[i-1]+nn;
    givenproj[i]=givenproj[i-1]+nn;
    svd[i]=svd[i-1]+nn;
  }

  /* Add dense constraints */
  for (i=0;i<nden;i++){
    MakeDense(&sdt->amats->rows[smw->dind[i]],sdt->wd,n);
    ForwSubst(sdt->sf,sdt->wd,smw->upd[i]);
    for (j=0; j<ndiag; j++){
      smw->upd[i][smw->ndInd[j]] = -(smw->upd[i][smw->ndInd[j]]);
    }
  }
  
  for (i=0;i<mat2size;i++){
    for(j=i;j<mat2size;j++){
      if (i<nden){
        if (j<nden){
	  dtemp=VDotDD(smw->upd[i],smw->upd[j],n);
	  smw->mat2[i][j-i] = dtemp;
        }
	else if (j<nden+ndiag){
	  dtemp = smw->upd[i][smw->ndInd[j-nden]];
	  smw->mat2[i][j-i] = dtemp; 
	}
	else {smw->mat2[i][j-i]=0.0;}
      }
      else{
	if ((i==j)&&(j<nden+ndiag)) {smw->mat2[i][0] = 1.0;}
	else {smw->mat2[i][j-i] = 0.0;}
      }
      for (k=0; k<i ;k++){
        if (k<nden+ndiag){
          smw->mat2[i][j-i] = smw->mat2[i][j-i] - 
	    smw->mat2[k][i-k]*smw->mat2[k][j-k]/smw->mat2[k][0];
	}
      }
    } /* end for j */
  } /* end for i */


  /* Copy contents into upper triangular coefficient matrix */
  for (i=0; i<nn; i++){
    for (j=i; j<nn; j++){
      coeffmat[i][j]= smw->mat2[i][j-i]*smw->mat2[i][j-i];
    }
  }


  for (i=0;i<mat2size; i++){
    smw->v2[i] = 1.0;
  }
  for (i=0;i<nden+ndiag;i++){
    smw->v2[i] = sqrt( fabs(smw->mat2[i][0]) );
  }
 
  for (i=0;i<mat2size;i++){
    for (j=i;j<mat2size;j++){
      dtemp = 0.0;
      for (k=j;k<nden+ndiag;k++){
 	if (k<nden){
	  dtemp = dtemp +
	    smw->mat2[i][k-i]*smw->mat2[j][k-j]*(-coef[smw->dind[k]]);
	}
	else{
	  dtemp = dtemp + 
	          smw->mat2[i][k-i]*smw->mat2[j][k-j]*(-2);
	}
      }
      smw->mat2[i][j-i]=dtemp /(smw->v2[i]*smw->v2[j]);
    }
  }
  
  for (i=0; i<mat2size; i++){
    smw->mat2[i][0] = smw->mat2[i][0] + 1.0;
  }

  /* Copy matrix into the svd matrix */
  for (i=0; i<nn; i++){
    for (j=i; j<nn; j++){
      svd[i][j]= smw->mat2[i][j-i];
    }
  }
  for (i=0; i<nn; i++){
    for (j=i+1; j<nn; j++){
      svd[j][i]= smw->mat2[i][j-i];
    }
  }

  for (i=0; i<nn; i++){
    givenproj[i][i] = 1.0;
  }

  /* Apply Givens transformations */
  for (j=0; j<nn; j++){
    for (i=0; i<nn; i++){
      if (i !=j){
	ComputeGivens(svd[i][i], svd[i][j], &c, &s);
	PreMultGivens(i, j, c, s, svd,nn); 
	/* Then premultiply the projection matrix */
	PreMultGivens(i, j, c, s, givenproj,nn); 
      }
    }
  }  

  /* Take square root of the singular values and subtract identity  */
  for (i=0; i<nn; i++){
    svd[i][i] = sqrt (fabs(svd[i][i]) ) -1.0;
  }

  for (i=0; i<nn; i++){
    temp = 0;
    for (j=0; j<i; j++){
      temp=temp+givenproj[i][j]*givenproj[i][j];
    }
    temp = temp + givenproj[i][i]*svd[i][i]*givenproj[i][i];
    for (j=i+1; j<nn; j++){
      temp=temp + givenproj[i][j]*givenproj[i][j];
    }
    rhs[i]=temp;
  }

  for (i=0; i<nn; i++){
    smw->v2[i]=0.0;
  }

  /* Solve upper triangular system */
  for (i=nden+ndiag-1; i>=0; i--){
    smw->v2[i]=rhs[i];
    for (j=nn-1; j>i; j--){
      smw->v2[i]=smw->v2[i]-coeffmat[i][j]*smw->v2[j];
    }
    smw->v2[i] = smw->v2[i]/coeffmat[i][i];
  }

  /* Free memory */
  free(coeffmat[0]);
  free(coeffmat);
  free(givenproj[0]);
  free(givenproj);
  free(svd[0]);
  free(svd);
  free(rhs);

  return 0;
}



int PreMultGivens(int i, int j, double c, double s, double **mat, int n){

  double mii, mij;

  mii = mat[i][i];
  mij = mat[i][j];

  mat[i][i]= c*mii - s*mij;
  mat[i][j]= s*mii+c*mij;

  return 1;
}

int ComputeGivens(double a, double b, double *c, double *s){

  double t1;

  if (b==0){
    *c=1; *s=0;
  } else {  
    if ( fabs(b) > fabs(a) ){
      t1=-a/b; *s=1/sqrt(1+t1*t1); *c=(*s)*t1;
    }
    else{
      t1=-b/a; *c=1/sqrt(1+t1*t1); *s=(*c)*t1;
    }
  }
  return 0;
}




